"""
Constants and functions for computing geometry.

Attributes: Geometrical constants

    π (Quantity): The ratio of a circle's diameter to its radius
"""

import math

from measured import One

π = math.pi * One
